/* Copyright (C) 2004, 2005, 2006  Andreas Kunz
 * 
 * YAReG is free software; you can redistribute and/or modify it under the terms of the 
 * GNU General Public License (version 2 or later) as published by the Free Software Foundation.
 * This program is distributed WITHOUT ANY WARRANTY; See the GNU GPL for more details. 
 */

using System;
using System.Collections;
using System.Text.RegularExpressions;
using System.Windows.Forms;

namespace YAReG
{
	/// <summary>
	/// A PartitionNode is a special file system node with some partition information.
	/// </summary>
	public class PartitionNode : FsNode
	{
		// private member variables
		private PartType partitionType = PartType.UNDEFINED;
		private string partID = "";

		// public accessors
		public string PartID
		{
			get 
			{
				return partID; 
			}
		}

		public PartType PartitionType
		{
			get 
			{
				return partitionType; 
			}
		}


		/// <summary>
		/// Constructor creating a PartitionNode from raw command line tool output.
		/// </summary>
		/// <param name="line">raw data from a command line tool</param>
		/// <param name="source">identifier of the tool</param>
		public PartitionNode(string line, From source)
		{
			path = "";
			size = 0;
			attributes = "";
			nodeType = fsType.PARTITION;

			switch (source)
			{
				case From.RFSTOOL_PART:
					// parse rfstool output
					name		= Regex.Replace(line, @".*rive\s(\d*),\sPartition\s(\d*).*", "drive$1part$2");
					partition	= this;
					partitionType = PartType.REISERFS;
					partID		= Regex.Replace(line, @".*rive\s(\d*),\sPartition\s(\d*).*", "$1.$2");
					size		= Int64.Parse(Regex.Replace(line, @".*Size:\s(.*)\sBytes.*", "$1"));
					sizeString	= Regex.Replace(line, @".*Bytes\s(\(.*\)).*", "$1");
					nodeType	= fsType.PARTITION;
					this.Nodes.Add(FsNode.CreateDummy()); // append dummy node
					break;
				default:
					MessageBox.Show("Commandline tool data source " + 
						source + " unknown.", "Error");
					break;
			}
			Text = name;
			ImageIndex = (int) nodeType;
			files = new ArrayList();
		}
		

		/// <summary>
		/// Enumeration containing all the partition types YAReG understands.
		/// </summary>
		public enum PartType
		{
			UNDEFINED,
			REISERFS
		}
	}
}
