/* Copyright (C) 2004, 2005, 2006  Andreas Kunz
 * 
 * YAReG is free software; you can redistribute and/or modify it under the terms of the 
 * GNU General Public License (version 2 or later) as published by the Free Software Foundation.
 * This program is distributed WITHOUT ANY WARRANTY; See the GNU GPL for more details. 
 */

using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Configuration;
using System.Windows.Forms;
using System.Data;
using System.IO;
using System.Diagnostics;
using System.Text.RegularExpressions;
using System.Threading;

namespace YAReG
{
	/// <summary>
	/// The YAReG main window
	/// </summary>
	public class YAReGForm : System.Windows.Forms.Form
	{
		/************************************************************************/
		/* Change version number and title bar string HERE!                     */
		public static string gVersion     = "1.0"; 
		public static string gReleaseDate = "March 7th, 2006"; 
		public static string gTitleString = "YAReG - Yet Another Re(iserFStool) GUI " + gVersion;
		/************************************************************************/

		internal static string RFSTOOL = "rfstool.exe";
		private FsNode gCurrentNode;
		private string gTmpDir;
		

		#region generated members
		private System.Windows.Forms.MenuItem menuItemAbout;
		private System.Windows.Forms.MenuItem menuFile;
		private System.Windows.Forms.MenuItem menuItemExit;
		private System.Windows.Forms.MenuItem menuHelp;
		private System.Windows.Forms.MainMenu mainMenu;
		private System.Windows.Forms.TreeView dirTree;
		private System.Windows.Forms.ListView dirList;
		private System.Windows.Forms.StatusBar statusBar;
		private System.Windows.Forms.ColumnHeader columnName;
		private System.Windows.Forms.ColumnHeader columnSize;
		private System.Windows.Forms.ColumnHeader columnDate;
		private System.Windows.Forms.ColumnHeader columnAttributes;
		private System.Windows.Forms.ImageList imlSmallIcons;
		private System.Windows.Forms.ImageList imlLargeIcons;
		private System.Windows.Forms.MenuItem menuItem1;
		private System.Windows.Forms.MenuItem menuItemLargeIcons;
		private System.Windows.Forms.MenuItem menuItemSmallIcons;
		private System.Windows.Forms.MenuItem menuItemDetails;
		private System.Windows.Forms.MenuItem menuItemList;
		private System.Windows.Forms.StatusBarPanel statusBarPathPanel;
		private System.Windows.Forms.StatusBarPanel statusBarSizePanel;
		private System.Windows.Forms.StatusBarPanel statusBarSelectedPanel;
		private System.Windows.Forms.ContextMenu dirListcontextMenu;
		private System.Windows.Forms.MenuItem contextMenuCopyTo;
		private System.Windows.Forms.FolderBrowserDialog targetFolderDialog;
		private System.Windows.Forms.Splitter splitter;
		private System.ComponentModel.IContainer components;

#endregion

		/// <summary>
		/// Constructor (called by main())
		/// </summary>
		public YAReGForm()
		{
			// generated code
			InitializeComponent();

			// attach the sorting class to the dirList
			dirList.ListViewItemSorter = new DirListComparer();

			// populate the list of partitions
			loadPartitions();

			// try to read temp directory from config file
			string tmpRoot = ConfigurationSettings.AppSettings["TEMPDIR"];

			// if not successful...
			if (tmpRoot == null || tmpRoot == "" || !Directory.Exists(tmpRoot)) 
			{
				// ...create temp directory in system's TEMP directory...
				tmpRoot = Environment.GetEnvironmentVariable("TEMP");

				// ...or in the application directory
				if (tmpRoot == null || tmpRoot == "" || !Directory.Exists(tmpRoot))
					tmpRoot = Application.StartupPath;
			}

			if (!tmpRoot.EndsWith("\\"))
				tmpRoot += "\\";

			gTmpDir = tmpRoot + "YAReG-temp";
			try 
			{
				Directory.CreateDirectory(gTmpDir);
			} 
			catch (Exception e) 
			{
				MessageBox.Show("Exception when creating the temporary directory:" 
					+ "\n" + gTmpDir.ToString() + ":"
					+ "\n" + e.Message
					+ "\nExiting.",
					"Fatal Error");
				Application.Exit();
			}
		}


		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose( bool disposing )
		{
			if( disposing )
			{
				if (components != null) 
				{
					components.Dispose();
				}

				// delete the temp directory
				try 
				{
					Directory.Delete(gTmpDir, true);
				}
				catch 
				{
					// do nothing
				}
			}
			base.Dispose( disposing );
		}


		#region Windows Form Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
			this.components = new System.ComponentModel.Container();
			System.Resources.ResourceManager resources = new System.Resources.ResourceManager(typeof(YAReGForm));
			this.mainMenu = new System.Windows.Forms.MainMenu();
			this.menuFile = new System.Windows.Forms.MenuItem();
			this.menuItemExit = new System.Windows.Forms.MenuItem();
			this.menuItem1 = new System.Windows.Forms.MenuItem();
			this.menuItemList = new System.Windows.Forms.MenuItem();
			this.menuItemLargeIcons = new System.Windows.Forms.MenuItem();
			this.menuItemSmallIcons = new System.Windows.Forms.MenuItem();
			this.menuItemDetails = new System.Windows.Forms.MenuItem();
			this.menuHelp = new System.Windows.Forms.MenuItem();
			this.menuItemAbout = new System.Windows.Forms.MenuItem();
			this.dirTree = new System.Windows.Forms.TreeView();
			this.imlSmallIcons = new System.Windows.Forms.ImageList(this.components);
			this.dirList = new System.Windows.Forms.ListView();
			this.columnName = new System.Windows.Forms.ColumnHeader();
			this.columnSize = new System.Windows.Forms.ColumnHeader();
			this.columnDate = new System.Windows.Forms.ColumnHeader();
			this.columnAttributes = new System.Windows.Forms.ColumnHeader();
			this.imlLargeIcons = new System.Windows.Forms.ImageList(this.components);
			this.statusBar = new System.Windows.Forms.StatusBar();
			this.statusBarSelectedPanel = new System.Windows.Forms.StatusBarPanel();
			this.statusBarSizePanel = new System.Windows.Forms.StatusBarPanel();
			this.statusBarPathPanel = new System.Windows.Forms.StatusBarPanel();
			this.dirListcontextMenu = new System.Windows.Forms.ContextMenu();
			this.contextMenuCopyTo = new System.Windows.Forms.MenuItem();
			this.targetFolderDialog = new System.Windows.Forms.FolderBrowserDialog();
			this.splitter = new System.Windows.Forms.Splitter();
			((System.ComponentModel.ISupportInitialize)(this.statusBarSelectedPanel)).BeginInit();
			((System.ComponentModel.ISupportInitialize)(this.statusBarSizePanel)).BeginInit();
			((System.ComponentModel.ISupportInitialize)(this.statusBarPathPanel)).BeginInit();
			this.SuspendLayout();
			// 
			// mainMenu
			// 
			this.mainMenu.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
																					 this.menuFile,
																					 this.menuItem1,
																					 this.menuHelp});
			// 
			// menuFile
			// 
			this.menuFile.Index = 0;
			this.menuFile.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
																					 this.menuItemExit});
			this.menuFile.Text = "&File";
			// 
			// menuItemExit
			// 
			this.menuItemExit.Index = 0;
			this.menuItemExit.Text = "E&xit";
			this.menuItemExit.Click += new System.EventHandler(this.menuItemExit_Click);
			// 
			// menuItem1
			// 
			this.menuItem1.Index = 1;
			this.menuItem1.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
																					  this.menuItemList,
																					  this.menuItemLargeIcons,
																					  this.menuItemSmallIcons,
																					  this.menuItemDetails});
			this.menuItem1.Text = "&View";
			// 
			// menuItemList
			// 
			this.menuItemList.Index = 0;
			this.menuItemList.RadioCheck = true;
			this.menuItemList.Text = "List";
			this.menuItemList.Click += new System.EventHandler(this.menuItemList_Click);
			// 
			// menuItemLargeIcons
			// 
			this.menuItemLargeIcons.Index = 1;
			this.menuItemLargeIcons.RadioCheck = true;
			this.menuItemLargeIcons.Text = "Large Icons";
			this.menuItemLargeIcons.Click += new System.EventHandler(this.menuItemLargeIcons_Click);
			// 
			// menuItemSmallIcons
			// 
			this.menuItemSmallIcons.Index = 2;
			this.menuItemSmallIcons.RadioCheck = true;
			this.menuItemSmallIcons.Text = "Small Icons";
			this.menuItemSmallIcons.Click += new System.EventHandler(this.menuItemSmallIcons_Click);
			// 
			// menuItemDetails
			// 
			this.menuItemDetails.Checked = true;
			this.menuItemDetails.Index = 3;
			this.menuItemDetails.RadioCheck = true;
			this.menuItemDetails.Text = "Details";
			this.menuItemDetails.Click += new System.EventHandler(this.menuItemDetails_Click);
			// 
			// menuHelp
			// 
			this.menuHelp.Index = 2;
			this.menuHelp.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
																					 this.menuItemAbout});
			this.menuHelp.Text = "&Help";
			// 
			// menuItemAbout
			// 
			this.menuItemAbout.Index = 0;
			this.menuItemAbout.Text = "&About";
			this.menuItemAbout.Click += new System.EventHandler(this.menuItemAbout_Click);
			// 
			// dirTree
			// 
			this.dirTree.Dock = System.Windows.Forms.DockStyle.Left;
			this.dirTree.ImageList = this.imlSmallIcons;
			this.dirTree.Location = new System.Drawing.Point(0, 0);
			this.dirTree.Name = "dirTree";
			this.dirTree.Size = new System.Drawing.Size(206, 325);
			this.dirTree.TabIndex = 1;
			this.dirTree.AfterSelect += new System.Windows.Forms.TreeViewEventHandler(this.dirTree_AfterSelect);
			this.dirTree.BeforeExpand += new System.Windows.Forms.TreeViewCancelEventHandler(this.dirTree_BeforeExpand);
			// 
			// imlSmallIcons
			// 
			this.imlSmallIcons.ColorDepth = System.Windows.Forms.ColorDepth.Depth16Bit;
			this.imlSmallIcons.ImageSize = new System.Drawing.Size(16, 16);
			this.imlSmallIcons.ImageStream = ((System.Windows.Forms.ImageListStreamer)(resources.GetObject("imlSmallIcons.ImageStream")));
			this.imlSmallIcons.TransparentColor = System.Drawing.Color.Transparent;
			// 
			// dirList
			// 
			this.dirList.AllowColumnReorder = true;
			this.dirList.Columns.AddRange(new System.Windows.Forms.ColumnHeader[] {
																					  this.columnName,
																					  this.columnSize,
																					  this.columnDate,
																					  this.columnAttributes});
			this.dirList.Dock = System.Windows.Forms.DockStyle.Fill;
			this.dirList.LargeImageList = this.imlLargeIcons;
			this.dirList.Location = new System.Drawing.Point(209, 0);
			this.dirList.Name = "dirList";
			this.dirList.Size = new System.Drawing.Size(455, 325);
			this.dirList.SmallImageList = this.imlSmallIcons;
			this.dirList.TabIndex = 2;
			this.dirList.View = System.Windows.Forms.View.Details;
			this.dirList.MouseDown += new System.Windows.Forms.MouseEventHandler(this.dirList_MouseDown);
			this.dirList.ItemActivate += new System.EventHandler(this.dirList_ItemActivate);
			this.dirList.KeyPress += new System.Windows.Forms.KeyPressEventHandler(this.dirList_KeyPress);
			this.dirList.ColumnClick += new System.Windows.Forms.ColumnClickEventHandler(this.dirList_ColumnClick);
			this.dirList.ItemDrag += new System.Windows.Forms.ItemDragEventHandler(this.dirList_ItemDrag);
			this.dirList.SelectedIndexChanged += new System.EventHandler(this.dirList_SelectedIndexChanged);
			// 
			// columnName
			// 
			this.columnName.Text = "Name";
			this.columnName.Width = 177;
			// 
			// columnSize
			// 
			this.columnSize.Text = "Size";
			this.columnSize.TextAlign = System.Windows.Forms.HorizontalAlignment.Right;
			this.columnSize.Width = 73;
			// 
			// columnDate
			// 
			this.columnDate.Text = "Modified";
			this.columnDate.Width = 107;
			// 
			// columnAttributes
			// 
			this.columnAttributes.Text = "Attributes";
			this.columnAttributes.Width = 81;
			// 
			// imlLargeIcons
			// 
			this.imlLargeIcons.ColorDepth = System.Windows.Forms.ColorDepth.Depth32Bit;
			this.imlLargeIcons.ImageSize = new System.Drawing.Size(32, 32);
			this.imlLargeIcons.ImageStream = ((System.Windows.Forms.ImageListStreamer)(resources.GetObject("imlLargeIcons.ImageStream")));
			this.imlLargeIcons.TransparentColor = System.Drawing.Color.Transparent;
			// 
			// statusBar
			// 
			this.statusBar.Font = new System.Drawing.Font("Microsoft Sans Serif", 8.25F, System.Drawing.FontStyle.Bold, System.Drawing.GraphicsUnit.Point, ((System.Byte)(0)));
			this.statusBar.Location = new System.Drawing.Point(0, 325);
			this.statusBar.Name = "statusBar";
			this.statusBar.Panels.AddRange(new System.Windows.Forms.StatusBarPanel[] {
																						 this.statusBarSelectedPanel,
																						 this.statusBarSizePanel,
																						 this.statusBarPathPanel});
			this.statusBar.ShowPanels = true;
			this.statusBar.Size = new System.Drawing.Size(664, 20);
			this.statusBar.TabIndex = 3;
			// 
			// statusBarSelectedPanel
			// 
			this.statusBarSelectedPanel.AutoSize = System.Windows.Forms.StatusBarPanelAutoSize.Spring;
			this.statusBarSelectedPanel.MinWidth = 30;
			this.statusBarSelectedPanel.Width = 279;
			// 
			// statusBarSizePanel
			// 
			this.statusBarSizePanel.MinWidth = 30;
			this.statusBarSizePanel.Width = 91;
			// 
			// statusBarPathPanel
			// 
			this.statusBarPathPanel.MinWidth = 20;
			this.statusBarPathPanel.Width = 278;
			// 
			// dirListcontextMenu
			// 
			this.dirListcontextMenu.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
																							   this.contextMenuCopyTo});
			// 
			// contextMenuCopyTo
			// 
			this.contextMenuCopyTo.Index = 0;
			this.contextMenuCopyTo.Text = "Copy To..";
			this.contextMenuCopyTo.Click += new System.EventHandler(this.contextMenuCopyTo_Click);
			// 
			// targetFolderDialog
			// 
			this.targetFolderDialog.Description = "Please select the target folder for the file(s) to copy from the ReiserFS partiti" +
				"on.";
			// 
			// splitter
			// 
			this.splitter.Location = new System.Drawing.Point(206, 0);
			this.splitter.Name = "splitter";
			this.splitter.Size = new System.Drawing.Size(3, 325);
			this.splitter.TabIndex = 4;
			this.splitter.TabStop = false;
			// 
			// YAReGForm
			// 
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.ClientSize = new System.Drawing.Size(664, 345);
			this.Controls.Add(this.dirList);
			this.Controls.Add(this.splitter);
			this.Controls.Add(this.dirTree);
			this.Controls.Add(this.statusBar);
			this.Icon = ((System.Drawing.Icon)(resources.GetObject("$this.Icon")));
			this.Menu = this.mainMenu;
			this.Name = "YAReGForm";
			this.Text = "YAReG - Yet Another Re(iserfstool) GUI";
			((System.ComponentModel.ISupportInitialize)(this.statusBarSelectedPanel)).EndInit();
			((System.ComponentModel.ISupportInitialize)(this.statusBarSizePanel)).EndInit();
			((System.ComponentModel.ISupportInitialize)(this.statusBarPathPanel)).EndInit();
			this.ResumeLayout(false);

		}
		#endregion

		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main() 
		{
			// Before even opening the YAReG window:
			// check for rfstool, display an alert and exit if it is not found
			if (!File.Exists(Application.StartupPath + @"\rfstool.exe")) 
			{
				MessageBox.Show("rfstool.exe not found in the same directory as YAReG.exe."
					+ "\nYou need to download it from Gerson Kurz's web page in order to use YAReG.exe."
					+ "\n(currently: http://p-nand-q.com/download/rfstool/rfstool-0.14.zip)",
					"Fatal Error");
				Application.Exit();
			} 
			else 
			{ 
				Application.Run(new YAReGForm());
			}
		}


		/// <summary>
		/// Called by the constructor, this calls the underlying tool(s) to get a list of all
		/// supported Linux partitions.
		/// Insert calls for any added file system access methods HERE!
		/// </summary>
		private void loadPartitions()  
		{
			loadReiserFSPartitions();	// load reiserFS partitions using rfstool 
			//  ... call methods that use other tools for ext2/3 or other partition types
		}


		/// <summary>
		/// Calls rfstool and adds returned reiserFS partitions to the tree.
		/// (used during initialization)
		/// </summary>
		private void loadReiserFSPartitions()
		{
			Process p = new Process();
			p.StartInfo.UseShellExecute = false;
			p.StartInfo.RedirectStandardOutput = true;
			p.StartInfo.FileName = RFSTOOL;
			p.StartInfo.Arguments = "autodetect";
			p.StartInfo.CreateNoWindow = true;
			try 
			{
				p.Start();
				StreamReader output = p.StandardOutput;

				string line1 = "", line2 = "";
				FsNode partNode;
				dirTree.Nodes.Clear();
				while (output.Peek() > -1) 
				{
					line1 = output.ReadLine();
					//     MessageBox.Show(output.Peek().ToString());
					// This demonstrates a possible .NET bug:
					// After reading the first line of output, output.Peek() seems to 
					// return -1 even if there is still data in the stream.
					// This may lead to a YAReG bug when the first line(s) in the stream is/are
					// error message(s) e.g. about a drive not being readable:
					// line2 is not read and (output.Peek() > -1) returns false, thus ending
					// the loop and preventing any partitions from being detected.
					// First doing p.WaitForExit(); and then reading its data does solve the 
					// problem but might cause a hang with too much data.
					if (Regex.IsMatch(line1, "is a ReiserFS"))
					{
						line2 = output.ReadLine();
						partNode = new PartitionNode(line1 + line2, FsNode.From.RFSTOOL_PART);
						dirTree.Nodes.Add(partNode);
					}
				}
				p.WaitForExit();
			} 
			catch (Exception e) 
			{
				MessageBox.Show("An exception occurred while calling rfstool to detect the " 
					+ "ReiserFS partitions: "
					+ e.Message 
					+ "\nYou can go on using YAReG, but some partitions might not be listed.", 
					"Error");
			}
		}


		/// <summary>
		/// Is called whenever the currently active directory is changed by an 
		/// user action. If this new directory is not yet populated it does so.
		/// Anyway, it updates the tree and list view and various other UI 
		/// elements like the status bar to reflect the new directory.
		/// </summary>
		/// <param name="node">node changed to</param>
		private void changeToNode(FsNode node)
		{
			// make sure to not go to inexistent nodes. Rather go nowhere.
			if (node == null)
				return;

			// populate child nodes
			if (!node.IsPopulated 
				&& (node.NodeType == FsNode.fsType.DIRECTORY 
				    || node.NodeType == FsNode.fsType.PARTITION)) 
			{
				node.Populate();
			}

			// update dirTree
			dirTree.SelectedNode = node; // Does not trigger a Select event if the node
										 // is already selected. So there is no loop here.
			
			// update dirList
			dirList.Items.Clear();
			ListViewItem li;
			foreach (FsNode n in node.Nodes) // add directories 
			{
				li = new FsListItem(n);
				dirList.Items.Add(li);
			}
			foreach (FsNode n in node.Files) // add files
			{
				li = new FsListItem(n);
				dirList.Items.Add(li);
			} 

			// update current node + status bar/title text
			gCurrentNode = node;
			updateListSelection();
			statusBarPathPanel.Text = node.Path;
			if (node.NodeType == FsNode.fsType.DIRECTORY) 
			{
				statusBarPathPanel.Text += "/" + node.Name;
				// display the current folder name in the title bar (as Explorer does)
				this.Text = node.Name;
			}
			else
			{
				this.Text = gTitleString;
			}
		}


		/// <summary>
		/// Event handler for when the user selects a directory in the tree;
		/// changes the currently active directory to the selected one and triggers an
		/// update of the UI.
		/// </summary>
		/// <param name="sender">sender</param>
		/// <param name="e">event</param>
		private void dirTree_AfterSelect(object sender, System.Windows.Forms.TreeViewEventArgs e)
		{
			changeToNode((FsNode) dirTree.SelectedNode);
		}


		/// <summary>
		/// Event handler called when the user opens a closed folder or partition in the
		/// left-hand tree; populates the opened node if it was never populated before.
		/// </summary>
		/// <param name="sender">sender</param>
		/// <param name="e">event</param>
		private void dirTree_BeforeExpand(object sender, System.Windows.Forms.TreeViewCancelEventArgs e)
		{
			if (!((FsNode) e.Node).IsPopulated && 
					(((FsNode) e.Node).NodeType == FsNode.fsType.DIRECTORY ||
					 ((FsNode) e.Node).NodeType == FsNode.fsType.PARTITION))
				((FsNode) e.Node).Populate();
		}


		/// <summary>
		/// Event handler called when the user "opens" a file or directory in the right-hand
		/// file list; enters the directory/does NOT open the file (but shows an error message).
		/// </summary>
		/// <param name="sender">sender</param>
		/// <param name="e">event</param>
		private void dirList_ItemActivate(object sender, System.EventArgs e)
		{
			// Do nothing when user presses Return with nothing selected
			if (dirList.FocusedItem == null)
				return;

			FsNode nodeToEnter = ((FsNode) ((FsListItem) dirList.FocusedItem).Node);
			if (nodeToEnter.NodeType == FsNode.fsType.DIRECTORY)
			{                           // it is a directory, so let's enter
				changeToNode(nodeToEnter);
			} 
			else						// it is a regular file, do not open the file directly
			{
				MessageBox.Show("You cannot open files directly, please copy them to your Windows partition first.", 
								"Error"); // TODO:implement? Rather not because the user is going
										  // to think he can edit the file on the Linux partition
			}
		}


		/// <summary>
		/// Event handler called when the selection in the right-hand file list changes;
		/// triggers updating everything related.
		/// </summary>
		/// <param name="sender">sender</param>
		/// <param name="e">event</param>
		private void dirList_SelectedIndexChanged(object sender, System.EventArgs e)
		{
			updateListSelection();
		}


		/// <summary>
		/// Sets the status bar text according to the contents and selection of the right-hand
		/// file list.
		/// Currently called when the directory is changed 
		/// or the selection in the list of files is changed.
		/// </summary>
		private void updateListSelection()
		{
			long sumOfSizes = 0;
			if (dirList.SelectedItems.Count == 0) 
			{
				statusBarSelectedPanel.Text = dirList.Items.Count.ToString() +
					" Object(s)";
				foreach (FsListItem li in dirList.Items)
				{
					long size = ((FsListItem) li).Node.Size;
					if (size != -1)	  // do not count directories
						sumOfSizes += size;
				}
			}
			else
			{
				statusBarSelectedPanel.Text = dirList.SelectedItems.Count.ToString() + 
					" Object(s) selected";
				foreach (FsListItem li in dirList.SelectedItems)
				{
					long size = ((FsListItem) li).Node.Size;
					if (size != -1)	  // do not count directories
						sumOfSizes += size;
				}
			}

			if (sumOfSizes == 0)
				statusBarSizePanel.Text = "";
			else
				statusBarSizePanel.Text = statusSizeString(sumOfSizes);
		}


		/// <summary>
		/// Calculates a string describing the file size in a suitable human-
		/// readable format (like, "115 KB" for a file with size 117760 Bytes).
		/// Tries to mimic Windows Explorer's behaviour; not perfectly, though.
		/// </summary>
		/// <param name="size">file size in bytes</param>
		/// <returns>file size as string</returns>
		private string statusSizeString(long size)
		{
			if (size < 1024)
				return size.ToString() + " Byte";
			else 
			{
				float fSize = ((float) size) / 1024;
				if (fSize < 1024)
					return String.Format("{0:0.00} KB", fSize);
				else
					return String.Format("{0:0.00} MB", (fSize / 1024));
			}
		}


		#region menu items
		/// <summary>
		/// Invoked when the user clicks the "Help"->"About" menu entry.
		/// Displays version, copyright, web address and some system information.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void menuItemAbout_Click(object sender, System.EventArgs e)
		{
			MessageBox.Show("YAReG " + gVersion + "  (released " + gReleaseDate + ")"
				+ "\n\n(c) 2004, 2005, 2006  Andreas Kunz"
				+ "\nPublished under the terms of the GNU General Public License"
				+ "\n(you may redistribute YAReG under the GPL conditions)."
				+ "\nThis program comes with absolutely no warranty."
				+ "\n\nFurther information on:" 
				+ "\nhttp://yareg.akucom.de"
				+ "\n\n(using .NET version " + System.Environment.Version.ToString() + ")"
				+ "\n(using temporary directory " + gTmpDir + ")", 
				"About YAReG");
		}


		/// <summary>
		/// Invoked when the used selects the "File"->"Exit" menu item.
		/// Shuts down the application without asking. Asking sucks.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void menuItemExit_Click(object sender, System.EventArgs e)
		{
			Application.Exit();
		}


		/// <summary>
		/// Invoked when the user changes the list style in the "View" menu.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void menuItemList_Click(object sender, System.EventArgs e)
		{
			dirList.View = View.List;
			checkViewMenu(menuItemList);
		}


		/// <summary>
		/// Invoked when the user changes the list style in the "View" menu.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void menuItemSmallIcons_Click(object sender, System.EventArgs e)
		{
			dirList.View = View.SmallIcon;
			checkViewMenu(menuItemSmallIcons);
		}


		/// <summary>
		/// Invoked when the user changes the list style in the "View" menu.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void menuItemLargeIcons_Click(object sender, System.EventArgs e)
		{
			dirList.View = View.LargeIcon;
			checkViewMenu(menuItemLargeIcons);
		}


		/// <summary>
		/// Invoked when the user changes the list style in the "View" menu.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void menuItemDetails_Click(object sender, System.EventArgs e)
		{
			dirList.View = View.Details;
			checkViewMenu(menuItemDetails);
		}


		/// <summary>
		/// Deselects all entries for the different list styles in the "View" menu and
		/// selects the entry given as parameter.
		/// </summary>
		/// <param name="mItem"></param>
		private void checkViewMenu(MenuItem mItem)
		{
			menuItemList.Checked = false;
			menuItemSmallIcons.Checked = false;
			menuItemLargeIcons.Checked = false;
			menuItemDetails.Checked = false;
			mItem.Checked = true;
		}
		#endregion


		/// <summary>
		/// Called when the user starts a drag&drop action in the file list.
		/// For each selected file/directory rfstool is called to copy it to the temporary 
		/// directory.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void dirList_ItemDrag(object sender, System.Windows.Forms.ItemDragEventArgs e)
		{
			string draggedPath = ((FsListItem) dirList.SelectedItems[0]).Node.Path;

			// copy the files to the temp directory
			FsNode draggedNode;
			string sourceFileName;
			string targetFileName;
			string[] draggedFiles = new string[dirList.SelectedItems.Count];
            for (int i = 0; i < dirList.SelectedItems.Count; i++)
			{
				draggedNode = ((FsListItem) dirList.SelectedItems[i]).Node;

				// replace double quotes in source file names
				// this seems to work for files like 'foo"bar"', but not for 'foo"'
				sourceFileName = Regex.Replace(draggedNode.Name, "\"", "\\\"");

				// Let the user know what is happening
				statusBarSelectedPanel.Text = "Copying from ReiserFS partition: " + sourceFileName;
				this.Cursor = Cursors.WaitCursor;

				// replace special characters (that cannot be in Windows file names
				// but can be in Linux ones) with underscores for target files
				targetFileName = Regex.Replace(draggedNode.Name, @"[""\?\*\|\<\>\:\\\/]", "_");
				draggedFiles[i] = gTmpDir + @"\" + targetFileName;

				Process p = new Process();
				p.StartInfo.UseShellExecute = false;
				p.StartInfo.FileName = RFSTOOL;

				string args = "cp -p" + draggedNode.Partition.PartID + " " +
					"\"" + draggedPath + "/" + sourceFileName + "\"" + 
					" \"" +	gTmpDir + "\\" + targetFileName + "\""; 

				p.StartInfo.Arguments = args;
				p.StartInfo.CreateNoWindow = true;
				try 
				{
					p.Start();
					p.WaitForExit();
				}
				catch (Exception ex) 
				{
					MessageBox.Show("An exception occurred while calling rfstool to copy data from a ReiserFS partition: "
						+ ex.Message 
						+ "\nPlease inform the YAReG author about this problem.", 
						"Error");
				} 
			}

			// Let the user know that copying from ReiserFS partition is finished
			statusBarSelectedPanel.Text = "File(s) copied to TEMP, moving to target directory.";
			this.Cursor = Cursors.Default;

			// actually give the files in the temp directory to Windows for Drag&Drop
			// move the files from temp to target, so they are usually not left in temp
			// until YAReG is closed (only when drop is not executed). This way there is 
			// also the chance that the files do not have to be copied twice (if temp and
			// target are on same partition)
			DoDragDrop(new DndFiles(draggedFiles), DragDropEffects.Move);

			// reset status bar text
			updateListSelection();
		}


		/// <summary>
		/// Invoked when the user clicks on a column header.
		/// This sorts the list by the selected column.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void dirList_ColumnClick(object sender, System.Windows.Forms.ColumnClickEventArgs e)
		{
			// get the sorter
			DirListComparer sorter = (DirListComparer) dirList.ListViewItemSorter;

			// is this column already sorted?
			if (e.Column == sorter.ColumnToSort)
			{
				// Yes: reverse sort order, leave column the same
				if (sorter.Order == SortOrder.Ascending)
					sorter.Order = SortOrder.Descending;
				else
					sorter.Order = SortOrder.Ascending;
			}
			else
			{
				// set sort column, sort ascending
				sorter.ColumnToSort = e.Column;
				sorter.Order = SortOrder.Ascending;
			}

			// now sort using the above settings
			dirList.Sort();
		}


		/// <summary>
		/// Catches right-clicks on the dirList and displays the context menu.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void dirList_MouseDown(object sender, System.Windows.Forms.MouseEventArgs e)
		{
			// only handle right clicks
			if (e.Button != MouseButtons.Right)
				return;

			// if the list item below the mouse is not selected: select it
			ListViewItem target = dirList.GetItemAt(e.X, e.Y);

			// if user did not click on any item: clear selection and return
			if (target == null) 
			{
				dirList.SelectedItems.Clear();
				return;
			}

			// if there are items selected but the used right-clicked on a different item
			// deselect the others and select this item
			if (dirList.SelectedItems.Count == 0
				|| !target.Selected) 
			{
				dirList.SelectedItems.Clear();
				target.Selected = true;
			}

			// display the context menu
			dirListcontextMenu.Show(dirList, new Point(e.X, e.Y));
		}


		/// <summary>
		/// Invoked when the used selected "Copy To..." from the context menu.
		/// Prompts for the target folder and calls rfstool for every selected file/folder
		/// to transfer the data.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void contextMenuCopyTo_Click(object sender, System.EventArgs e)
		{
			// prompt user for target folder
			DialogResult result = targetFolderDialog.ShowDialog();

			// target selected, do some checks and transfer the files
			if (result == DialogResult.OK) 
			{
				string sourcePath = ((FsListItem) dirList.SelectedItems[0]).Node.Path;
				string targetFolder = targetFolderDialog.SelectedPath;

				// handle all the selected items
				ArrayList files = new ArrayList();
				foreach (object item in dirList.SelectedItems) 
				{
					FsNode draggedNode = ((FsListItem) item).Node;

					// replace double quotes in source file names
					// this seems to work for files like 'foo"bar"', but not for 'foo"'
					string sourceFileName = Regex.Replace(draggedNode.Name, "\"", "\\\"");
					string completeSourceName = "\"" + sourcePath + "/" + sourceFileName + "\"";

					// Let the user know what is happening
					statusBarSelectedPanel.Text = "Copying from ReiserFS partition: " + sourceFileName;

					// replace special characters (that cannot be in Windows file names
					// but can be in Linux ones) with underscores for target files
					string targetFileName = Regex.Replace(draggedNode.Name, @"[""\?\*\|\<\>\:\\\/]", "_");
					if (!targetFolder.EndsWith("\\"))
						targetFolder += "\\";
					string filePathName = targetFolder + targetFileName;
					string completeTargetName = "\"" + filePathName + "\"";


					// check whether targt file/directory already exists and ask/warn user
					if (File.Exists(filePathName)) 
					{
						result = MessageBox.Show("File " + filePathName 
							+ " already exists and will be overwritten (when it is not set read-only).\nDo you want to allow this?", 
							"Target File Exists", 
							MessageBoxButtons.YesNo,
							MessageBoxIcon.Warning,
							MessageBoxDefaultButton.Button2); 
						
						// skip to next file if user does not want to overwrite 
						if (result != DialogResult.Yes)
							continue;
					}
					if (Directory.Exists(filePathName)) 
					{
						result = MessageBox.Show("Directory " + filePathName 
							+ " already exists.\nAll files in that directory with the same name as a copied file will be overwritten (exception: read-only files)!" 
							+ "\nDo you want to allow this?", 
							"Target Directory Exists",
							MessageBoxButtons.YesNo,
							MessageBoxIcon.Warning,
							MessageBoxDefaultButton.Button2); 

						// skip to next file if user does not want to overwrite 
						if (result != DialogResult.Yes)
							continue;
					}

					// setup and start rfstool process
					Process p = new Process();
					p.StartInfo.UseShellExecute = false;
					p.StartInfo.FileName = RFSTOOL;

					string args = "cp -p" + draggedNode.Partition.PartID + " " 
						+ completeSourceName + " " 
						+ completeTargetName; 

					p.StartInfo.Arguments = args;
					p.StartInfo.CreateNoWindow = true;
					try 
					{
						p.Start();
						p.WaitForExit();
					}
					catch (Exception ex) 
					{
						MessageBox.Show("An exception occurred while calling rfstool to copy data from a ReiserFS partition: "
							+ ex.Message 
							+ "\nPlease inform the YAReG author about this problem.", 
							"Error");
					} 
				}

				// reset status bar text
				updateListSelection();
			}
		}

		private void dirList_KeyPress(object sender, System.Windows.Forms.KeyPressEventArgs e)
		{
			// "Backspace" => go up one directory (if possible)
			if (e.KeyChar.ToString().Equals("\b")) 
			{
				// do nothing when the current node is a partition (root node)
				if (gCurrentNode.NodeType != FsNode.fsType.PARTITION 
					&& (FsNode) gCurrentNode.Parent != null) 
				{
					changeToNode((FsNode) gCurrentNode.Parent);
				}
			}
		}
	}
}